from django.contrib.contenttypes.models import ContentType
from django.db import models

from six import python_2_unicode_compatible

from restraint.managers import PermSetManager, PermManager, PermLevelManager, PermAccessManager


@python_2_unicode_compatible
class PermSet(models.Model):
    """
    This is essentially a group that has a name.
    Each PermSet will have a PermAccess associated with it that will contain the groups full permissions
    The permissions are indicated through PermAccess.perm_levels.

    Fields:
     - Name
     - Display_name
     - Perm_access (One To One to PermAccess) For this set, this user type has these levels

    Example:
     - Name: manager
     - Display Name: Managers
    """
    name = models.CharField(max_length=256, unique=True, blank=True)
    display_name = models.TextField(blank=True)
    is_private = models.BooleanField(default=False)

    objects = PermSetManager()

    def __str__(self):
        return f'{self.display_name}:{self.name}'


@python_2_unicode_compatible
class Perm(models.Model):
    """
    This is the actual permission name specific to what each app will add, for example competition_all.

    Fields:
     - Name
     - Display_name

    Example:
     - Name: can_edit_accounts
     - Display Name: Users: Edit
     - Levels (Reverse Many To Many to PermLevel)
    """
    name = models.CharField(max_length=256, unique=True, blank=True)
    display_name = models.TextField(blank=True)

    objects = PermManager()

    def __str__(self):
        return f'{self.display_name}:{self.name}'


@python_2_unicode_compatible
class PermLevel(models.Model):
    """
    Each specific Perm can have different levels of access for the same permission.
    A blank level indicates that its either a yes/no decision
    Other options might look like all, created_by, etc...
    This should be unique and handled by each app independently

    Fields:
     - Perm (Foreign Key)
     - Name
     - Display_name

    Example:
    This is an example of a permission that can have multiple levels of permissions.
    Let's take the example permission of can_edit_accounts with three different permission levels:
     - Created_by: Can only edit accounts created by the user tied to the permission
     - Own: Can only edit their own account
     - Under: Can edit accounts that they manage

    Created By:
     - Permission: can_edit_accounts
     - Name: created_by
     - Display Name: Can Edit Users Created

    Own:
     - Permission: can_edit_accounts
     - Name: own
     - Display Name: Can Edit Own Account

    Under:
     - Permission: can_edit_accounts
     - Name: under
     - Display Name: Can Edit Subordinate Accounts
    """
    perm = models.ForeignKey(Perm, on_delete=models.CASCADE)
    name = models.CharField(max_length=256, blank=True)
    display_name = models.TextField(blank=True)

    objects = PermLevelManager()

    class Meta:
        unique_together = ('perm', 'name')

    def __str__(self):
        return f'{self.name}:{self.display_name}[PERM]{self.perm}'


class PermAccess(models.Model):
    """
    Provides access a list of permission levels for a permission set or for an individual
    user.

    This is what is used to determine what groups or individuals have access to.
    The access is determined by what levels are assigned to each group or individual.
    This can be found in the many to many relationship to levels `PermAccess.perm_levels.all()`

    Fields:
     - PermSet (One To One)
     - Perm_user_type (ContentType, Generic Foreign Key)
     - Perm_user_id (id of the object associated with the content type)
     - Perm_levels (Many to Many PermLevel)

    Example:
    An example would be for the manager permission set and allowing any manager to have access to two levels,
    own and under so they can edit their own account and any subordinate accounts.
     - Perm Set: manager
     - User Type: None
     - User Id: 0
     - Perm Levels
        - Own
        - Under
    """
    perm_set = models.OneToOneField(PermSet, null=True, default=None, on_delete=models.CASCADE)
    perm_user_type = models.ForeignKey(ContentType, null=True, default=None, on_delete=models.CASCADE)
    perm_user_id = models.PositiveIntegerField(default=0)
    perm_levels = models.ManyToManyField(PermLevel)

    class Meta:
        unique_together = ('perm_user_type', 'perm_user_id')

    objects = PermAccessManager()

    def __str__(self):  # pragma: no cover
        return f'[PERM_SET]{self.perm_set}[USER_TYPE]{self.perm_user_type}[USER_ID]{self.perm_user_id}'



# Generated by Django 3.2.12 on 2022-02-20 21:32

from django.db import migrations, models


def update_private_permission_sets(apps, schema_editor):
    PermSet = apps.get_model('restraint', 'PermSet')
    for perm_set in PermSet.objects.all():
        perm_set.is_private = True
        perm_set.save(update_fields=['is_private'])


class Migration(migrations.Migration):

    dependencies = [
        ('restraint', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='permset',
            name='is_private',
            field=models.BooleanField(default=False),
        ),
        migrations.RunPython(update_private_permission_sets)
    ]



from django.contrib.contenttypes.models import ContentType
from django.db import models
from manager_utils import sync


class PermSetManager(models.Manager):
    def sync_perm_sets(self, perm_sets):
        """
        Syncs all private perm sets the provided dictionary of perm sets to PermSet models.
        """
        from restraint.models import PermSet
        sync(
            queryset=self.get_queryset().filter(
                is_private=True
            ),
            model_objs=[
                PermSet(
                    name=name,
                    display_name=config.get('display_name', ''),
                    is_private=True
                )
                for name, config in perm_sets.items()
            ],
            unique_fields=[
                'name'
            ],
            update_fields=[
                'display_name'
            ]
        )


class PermManager(models.Manager):
    def sync_perms(self, perms):
        """
        Syncs the perms to Perm models.
        """
        from restraint.models import Perm
        return sync(
            queryset=self.get_queryset(),
            model_objs=[
                Perm(
                    name=name,
                    display_name=config.get('display_name', '')
                )
                for name, config in perms.items()
            ],
            unique_fields=[
                'name'
            ],
            update_fields=[
                'display_name'
            ],
            return_upserts_distinct=True
        )


class PermLevelManager(models.Manager):
    def sync_perm_levels(self, perms):
        """
        Given a dictionary of perms that map to perm levels, sync the perm levels
        to PermLevel objects in the database.
        """
        from restraint.models import Perm, PermLevel
        perm_objs = {
            p.name: p
            for p in Perm.objects.all()
        }
        perm_levels = []
        for perm, perm_config in perms.items():
            assert(perm_config['levels'])
            for level, level_config in perm_config['levels'].items():
                perm_levels.append(PermLevel(
                    perm=perm_objs[perm],
                    name=level,
                    display_name=level_config.get('display_name', '')
                ))
        sync(
            queryset=self.get_queryset(),
            model_objs=perm_levels,
            unique_fields=[
                'name',
                'perm'
            ],
            update_fields=[
                'display_name'
            ]
        )


class PermAccessManager(models.Manager):
    def set_default(self, permission_set_name, permission_name, levels=None):
        """
        Sets default levels for a permission for a permission set
        :param permission_set_name: The name of the permission set
        :param permission_name: The name of the permission
        :param levels: A list of levels
        """
        from restraint.models import PermSet, PermAccess, PermLevel
        permission_access = PermAccess.objects.get_or_create(
            perm_set=PermSet.objects.get(name=permission_set_name)
        )[0]
        if levels:
            permission_levels = PermLevel.objects.filter(
                perm__name=permission_name,
                name__in=levels
            )
            permission_access.perm_levels.set(permission_levels)
        else:
            permission_access.perm_levels.clear()

    def update_perm_set_access(self, config, new_perms=None, flush_previous_config=False):
        """
        Update the access for private perm sets with a config. The user can optionally flush
        the previous config and set it to the new one.
        """

        # Do model imports to avoid circular
        from restraint.models import PermSet, PermAccess, PermLevel

        # Ensure that new perms is not none
        if new_perms is None:  # pragma: no cover
            new_perms = []

        # Loop over each private permission set
        for perm_set in PermSet.objects.filter(is_private=True):
            perm_access, created = PermAccess.objects.get_or_create(perm_set=perm_set)
            perm_access_levels = []
            for perm, perm_levels in config.get(perm_set.name, {}).items():
                # If we are not flushing the previous config, continue if the perm not among the newly created perms
                # this is necessary because perm access is mutable; We don't want to destroy modifications made to
                # existing permissions
                if not created and not flush_previous_config and perm not in [p.name for p in new_perms]:
                    continue
                assert(perm_levels)
                perm_access_levels.extend(PermLevel.objects.filter(perm__name=perm, name__in=perm_levels))

            if flush_previous_config:
                perm_access.perm_levels.clear()
            perm_access.perm_levels.add(*perm_access_levels)

    def add_individual_access(self, user, perm_name, level_name):
        """
        Given a user, a permission name, and the name of the level, add the level in the permission access for
        the individual user.
        """
        from restraint.models import PermAccess, PermLevel
        pa, created = PermAccess.objects.get_or_create(
            perm_user_id=user.id,
            perm_user_type=ContentType.objects.get_for_model(user)
        )
        pa.perm_levels.add(PermLevel.objects.get(
            perm__name=perm_name,
            name=level_name
        ))

    def remove_individual_access(self, user, perm_name, level_name):
        """
        Given a user, a permission name, and the name of the level, remove the level in the permission access for
        the individual user.
        """
        from restraint.models import PermAccess, PermLevel
        pa = PermAccess.objects.get(
            perm_user_id=user.id,
            perm_user_type=ContentType.objects.get_for_model(user)
        )
        pa.perm_levels.remove(PermLevel.objects.get(
            perm__name=perm_name,
            name=level_name
        ))



from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.test import SimpleTestCase, TestCase
from django_dynamic_fixture import G
from mock import patch, Mock

from restraint import core, constants
from restraint.models import PermSet, Perm, PermLevel, PermAccess
import restraint.tests.configuration as test_configuration


class TestGetRestraintConfig(SimpleTestCase):
    def test_get_configuration(self):
        self.assertEqual(
            core.get_restraint_config()['perm_sets'],
            test_configuration.get_configuration()['perm_sets']
        )


class TestRestraintLoadPerms(TestCase):
    def setUp(self):
        core.update_restraint_db()

    def test_individual_user(self):
        # Create an individual user
        user = G(User, is_superuser=False, is_staff=False)

        # Get the users restraints
        restraints = core.Restraint(user)
        perms = restraints.perms
        self.assertEqual(
            perms,
            {
                'can_edit_stuff': {
                    'some_stuff': test_configuration.user_some_stuff_id_filter
                }
            }
        )

    def test_super_user(self):
        # Make a user that is a superuser and verify they get all proper permissions
        user = G(User, is_superuser=True)
        restraints = core.Restraint(user)
        perms = restraints.perms
        self.assertEquals(
            perms,
            {
                'can_edit_stuff': {
                    'all_stuff': None,
                    'some_stuff': test_configuration.user_some_stuff_id_filter,
                },
                'can_view_stuff': {
                    '': None
                },
                'can_access_users_named_foo': {
                    '': None
                }
            }
        )

    def test_user_with_additional_perms(self):
        # Make an individual user and verify they get all of the perms.
        # Also, add an individual permission to the user and verify they get that too
        user = G(User, is_superuser=False, is_staff=False)
        pa = G(PermAccess, perm_user_id=user.id, perm_user_type=ContentType.objects.get_for_model(user))
        pa.perm_levels.add(PermLevel.objects.get(name='all_stuff'))

        r = core.Restraint(user)
        perms = r.perms
        self.assertEquals(
            perms,
            {
                'can_edit_stuff': {
                    'some_stuff': test_configuration.user_some_stuff_id_filter,
                    'all_stuff': None
                }
            }
        )

    def test_load_some_perms(self):
        user = G(User, is_superuser=True)
        restraints = core.Restraint(user, ['can_edit_stuff'])
        perms = restraints.perms
        self.assertEquals(
            perms,
            {
                'can_edit_stuff': {
                    'all_stuff': None,
                    'some_stuff': test_configuration.user_some_stuff_id_filter
                }
            }
        )


class TestRestraintHasPerms(SimpleTestCase):
    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_w_level_true(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertTrue(r.has_perm('can_edit_stuff', 'all_stuff'))

    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_w_level_false(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertFalse(r.has_perm('can_edit_stuff', 'no_stuff'))

    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_wo_level_true(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertTrue(r.has_perm('can_edit_stuff'))

    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_wo_level_false(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertFalse(r.has_perm('can_mess_with_stuff'))


class TestRestraintFilterQSet(TestCase):
    def setUp(self):
        core.update_restraint_db()

    def test_filter_qset_global_access(self):
        # Make a user that is a superuser and verify they get all of the
        # super user perms
        u = G(User, is_superuser=True)
        # Make another user that they will be able to edit
        u2 = G(User)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([u, u2]))

    def test_filter_qset_local_access(self):
        # Make a user that is not a superuser
        u = G(User, is_superuser=False)
        # Make another user that they will not be able to edit
        G(User)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([u]))

    def test_filter_qset_multiple_local_access(self):
        # Make a user that is staff
        u = G(User, is_superuser=False, is_staff=True)
        # Make another user that they will not be able to edit
        G(User)
        # Make another super user that they will be able to edit
        u2 = G(User, is_superuser=True)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([u, u2]))

    def test_filter_qset_no_perms(self):
        # Make a user that is staff
        u = G(User, is_superuser=False, is_staff=True)
        # Load permissions that will not give them access to edit any accounts
        r = core.Restraint(u, ['bad_perm'])

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([]))

    def test_filter_qset_restrict_subset(self):
        models = [
            G(User, first_name='foo', last_name='foofington'),
            G(User, first_name='bar', last_name='barski'),
            G(User, first_name='foo', last_name='foogeelala'),
        ]
        # Make a user that is a superuser and verify they get all of the
        # super user perms
        u = G(User, is_superuser=True)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(
            User.objects.all(), 'can_access_users_named_foo', restrict_kwargs={'first_name': 'foo'})
        self.assertEquals(set(filtered_qset), set(models + [u]))

    def test_filter_qset_restrict_subset_no_perms(self):
        models = [
            G(User, first_name='foo', last_name='foofington'),
            G(User, first_name='bar', last_name='barski'),
            G(User, first_name='foo', last_name='foogeelala'),
        ]
        # Make a user that is a superuser and verify they get all of the
        # super user perms
        u = G(User, is_superuser=False)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(
            User.objects.all(), 'can_access_users_named_foo', restrict_kwargs={'first_name': 'foo'})
        self.assertEquals(set(filtered_qset), set([models[1]] + [u]))


class UpdateRestraintDbTest(TestCase):
    def add_custom_permission_set(self):
        # Setup a custom permission set
        custom_permission_set = PermSet.objects.create(
            name='custom',
            display_name='Custom',
            is_private=False
        )

        # Add some custom default access levels
        PermAccess.objects.set_default(
            permission_set_name=custom_permission_set.name,
            permission_name='can_edit_stuff',
            levels=['all_stuff']
        )

    @patch.object(core, 'get_restraint_config')
    def test_full_update_scenario_not_flush_default_access(self, mock_get_restraint_config):
        mock_get_restraint_config.return_value = {
            'perm_sets': {
                'global': {
                    'display_name': 'Global',
                },
                'restricted': {
                    'display_name': 'Restricted',
                },
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                },
            },
            'default_access': {
                'global': {
                    'can_edit_stuff': ['all_stuff', 'some_stuff'],
                    'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                },
                'restricted': {
                    'can_edit_stuff': ['some_stuff'],
                }
            }
        }
        core.update_restraint_db()
        core.update_restraint_db()

    @patch.object(core, 'get_restraint_config')
    def test_full_update_scenario_not_flush_default_access_update_new_perm(self, mock_get_restraint_config):
        """
        Verifies that existing permission set is given access to new permission
        """
        config = {
            'perm_sets': {
                'global': {
                    'display_name': 'Global',
                },
                'restricted': {
                    'display_name': 'Restricted',
                },
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                },
            },
            'default_access': {
                'global': {
                    'can_edit_stuff': ['all_stuff', 'some_stuff'],
                    'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                },
                'restricted': {
                    'can_edit_stuff': ['some_stuff'],
                }
            }
        }
        mock_get_restraint_config.return_value = config
        core.update_restraint_db()
        self.add_custom_permission_set()

        # add permission
        config['perms']['can_do_stuff'] = {
            'display_name': 'Can Do Stuff',
            'levels': {
                'all_stuff': {
                    'display_name': 'All Stuff',
                    'id_filter': None,
                },
                'this_thing': {
                    'display_name': 'This Thing',
                    'id_filter': None,
                },
            },
        }
        config['perms']['can_alter_stuff'] = {
            'display_name': 'Can Alter Stuff',
            'levels': constants.BOOLEAN_LEVELS_CONFIG
        }
        config['default_access']['global']['can_alter_stuff'] = [constants.BOOLEAN_LEVELS_NAME]
        config['default_access']['restricted']['can_do_stuff'] = ['all_stuff', 'this_thing']
        mock_get_restraint_config.return_value = config

        # update again
        core.update_restraint_db()

        self.assertEquals(
            set(PermSet.objects.filter(is_private=True).values_list('name', flat=True)),
            {'global', 'restricted'}
        )
        self.assertEquals(
            set(PermSet.objects.all().values_list('name', flat=True)),
            {'global', 'restricted', 'custom'}
        )

        self.assertEquals(
            set(Perm.objects.values_list('name', flat=True)),
            {'can_view_stuff', 'can_edit_stuff', 'can_do_stuff', 'can_alter_stuff'}
        )

        self.assertEquals(
            list(PermLevel.objects.order_by(
                'perm__name',
                'name'
            ).values_list(
                'perm__name',
                'name'
            )),
            [
                ('can_alter_stuff', ''),
                ('can_do_stuff', 'all_stuff'),
                ('can_do_stuff', 'this_thing'),
                ('can_edit_stuff', 'all_stuff'),
                ('can_edit_stuff', 'some_stuff'),
                ('can_view_stuff', '')
            ]
        )

        self.assertEquals(
            list(
                PermAccess.objects.order_by(
                    'perm_set__name',
                    'perm_levels__perm__name',
                    'perm_levels__name'
                ).values_list(
                    'perm_set__name',
                    'perm_levels__perm__name',
                    'perm_levels__name'
                )
            ),
            [
                ('custom', 'can_edit_stuff', 'all_stuff'),
                ('global', 'can_alter_stuff', ''),
                ('global', 'can_edit_stuff', 'all_stuff'),
                ('global', 'can_edit_stuff', 'some_stuff'),
                ('global', 'can_view_stuff', ''),
                ('restricted', 'can_do_stuff', 'all_stuff'),
                ('restricted', 'can_do_stuff', 'this_thing'),
                ('restricted', 'can_edit_stuff', 'some_stuff')
            ]
        )

    @patch.object(core, 'get_restraint_config')
    def test_full_update_scenario_flush_default_access(self, mock_get_restraint_config):
        config = {
            'perm_sets': {
                'global': {
                    'display_name': 'Global',
                },
                'restricted': {
                    'display_name': 'Restricted',
                },
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                },
            },
            'default_access': {
                'global': {
                    'can_edit_stuff': ['all_stuff', 'some_stuff'],
                    'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                },
                'restricted': {
                    'can_edit_stuff': ['some_stuff'],
                }
            }
        }
        mock_get_restraint_config.return_value = config
        core.update_restraint_db()
        self.add_custom_permission_set()

        config = {
            'perm_sets': {
                'global': {
                    'display_name': 'Global',
                },
                'restricted': {
                    'display_name': 'Restricted',
                },
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                },
            },
            'default_access': {
                'global': {
                    'can_edit_stuff': ['all_stuff'],
                },
            }
        }
        mock_get_restraint_config.return_value = config
        core.update_restraint_db(flush_default_access=True)

        self.assertEquals(
            set(PermSet.objects.filter(is_private=True).values_list('name', flat=True)),
            {'global', 'restricted'}
        )
        self.assertEquals(
            set(PermSet.objects.all().values_list('name', flat=True)),
            {'custom', 'global', 'restricted'}
        )

        self.assertEquals(
            set(Perm.objects.values_list('name', flat=True)),
            {'can_view_stuff', 'can_edit_stuff'}
        )

        self.assertEquals(
            list(PermLevel.objects.order_by(
                'perm__name',
                'name'
            ).values_list(
                'perm__name',
                'name'
            )),
            [
                ('can_edit_stuff', 'all_stuff'),
                ('can_edit_stuff', 'some_stuff'),
                ('can_view_stuff', '')
            ]
        )

        self.assertEquals(
            list(
                PermAccess.objects.order_by(
                    'perm_set__name',
                    'perm_levels__perm__name',
                    'perm_levels__name'
                ).values_list(
                    'perm_set__name',
                    'perm_levels__perm__name',
                    'perm_levels__name'
                )
            ),
            [
                ('custom', 'can_edit_stuff', 'all_stuff'),
                ('global', 'can_edit_stuff', 'all_stuff'),
                ('restricted', None, None)
            ]
        )



from django.contrib.auth.models import User

from restraint import constants


def perm_set_getter(u):
    perm_sets = ['individual']
    if u.is_superuser:
        perm_sets.append('super')
    if u.is_staff:
        perm_sets.append('staff')
    return perm_sets


def user_some_stuff_id_filter(user):
    return User.objects.filter(id=user.id).values_list('id', flat=True)


def user_only_super_users_id_filter(user):
    return User.objects.filter(is_superuser=True).values_list('id', flat=True)


def get_configuration():
    return {
        'perm_set_getter': perm_set_getter,
        'perm_sets': {
            'super': {
                'display_name': 'Super',
            },
            'individual': {
                'display_name': 'Individual',
            },
            'staff': {
                'display_name': 'Staff'
            }
        },
        'perms': {
            'can_edit_stuff': {
                'display_name': 'Can Edit Stuff',
                'levels': {
                    'all_stuff': {
                        'display_name': 'All Stuff',
                        'id_filter': None,
                    },
                    'some_stuff': {
                        'display_name': 'Some Stuff',
                        'id_filter': user_some_stuff_id_filter,
                    },
                    'only_superusers': {
                        'display_name': 'Only Superusers',
                        'id_filter': user_only_super_users_id_filter,
                    },
                },
            },
            'can_view_stuff': {
                'display_name': 'Can View Stuff',
                'levels': constants.BOOLEAN_LEVELS_CONFIG,
            },
            'can_access_users_named_foo': {
                'display_name': 'Can Foo',
                'levels': constants.BOOLEAN_LEVELS_CONFIG,
            }
        },
        'default_access': {
            'super': {
                'can_edit_stuff': ['all_stuff', 'some_stuff'],
                'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                'can_access_users_named_foo': [constants.BOOLEAN_LEVELS_NAME],
            },
            'individual': {
                'can_edit_stuff': ['some_stuff'],
            },
            'staff': {
                'can_edit_stuff': ['some_stuff', 'only_superusers']
            }
        }
    }
